#ifndef UFORTH_H
#define UFORTH_H
/*
  License for uforth 0.1 and later versions

  Copyright  2009 Todd Coram, todd@maplefish.com, USA.
  
  Permission is hereby granted, free of charge, to any person obtaining
  a copy of this software and associated documentation files (the
  "Software"), to deal in the Software without restriction, including
  without limitation the rights to use, copy, modify, merge, publish,
  distribute, sublicense, and/or sell copies of the Software, and to
  permit persons to whom the Software is furnished to do so, subject to
  the following conditions:
  
  The above copyright notice and this permission notice shall be
  included in all copies or substantial portions of the Software.
  
  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
  OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
  WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#include <stdint.h>

#define UFORTH_VERSION "0.3"
#define DICT_VERSION 0001

typedef int16_t CELL;
typedef char bool;
typedef enum { OK, E_NOT_A_WORD, E_STACK_UNDERFLOW, E_RSTACK_OVERFLOW,
	       E_DSTACK_OVERFLOW, 
	       E_MISSING_THEN,
	       E_MISSING_LOOP, E_ABORT } uforth_stat;

#define BYTES_PER_CELL sizeof(CELL)

#include "config.h"

/*
  iram -> 14 bytes + DS_CELLS*2 + RS_CELLS*2
*/
struct iram {
  char* inbufptr;		/* input buffer for interpreter */
  char* currword;		/* point to current word in inbufptr */
  CELL varidx;			/* keep track of next variable slot (neg #) */
  CELL didx;			/* data stack index */
  CELL ridx;			/* return stack index */
  CELL ds[DS_CELLS];		/* data stack */
  CELL rs[RS_CELLS];		/* return stack */
  uint8_t currwordlen;		/* length of current word in inbufptr */
  uint8_t compiling;		/* 0=interpreting, 1=colondef, 2=compiling */
  uint8_t ifnest;		/* if nesting counter (for compiler) */
  uint8_t donest;		/* do loop nesting counter (for compiler) */
};

struct dict {
  CELL version;			/* dictionary version number */
  CELL here;			/* top of dictionary */
  CELL last_word_idx;		/* top word's code token (used for searches) */
  CELL d[MAX_DICT_CELLS];	/* dictionary */
};


#define dpush(w) if (iram.didx == (DS_CELLS-1)) return E_DSTACK_OVERFLOW; \
  else iram.ds[++iram.didx] = w
#define dpop() iram.ds[iram.didx--]
#define dtop() iram.ds[iram.didx]

#define dpop32() ((uint32_t)dpop() << 16 | dpop())
#define dpush32(w2) { dpush(w2 & 0xFFFF); dpush(w2 >> 16); }

#define rpush(w) if (iram.ridx == (RS_CELLS-1)) return E_RSTACK_OVERFLOW; \
  else iram.rs[++iram.ridx] = w
#define rpop() iram.rs[iram.ridx--]
#define rtop() iram.rs[iram.ridx]


extern void uforth_init(void);
extern void uforth_abort(void);
extern uforth_stat uforth_interpret(char*);
extern uforth_stat c_handle(void);
char* uforth_next_word (void);

extern struct iram iram;
extern struct dict CONST *dict;

#ifdef USE_LITTLE_ENDIAN
# define BYTEPACK_FIRST(b) b
# define BYTEPACK_SECOND(b) ((CELL)b)<<8
#else
# define BYTEPACK_FIRST(b) ((CELL)b)<<8
# define BYTEPACK_SECOND(b) b
#endif

#ifdef BOOTSTRAP
# define INITIAL_DICT {DICT_VERSION, 0, -1, {0}}
#else
# include "core_dict.h"
#endif

/*
 Application hooks.
*/
void app_init(void);
void app_create_cmds(void);
uforth_stat app_handle_cmds(CELL n);

#endif	/* UFORTH_H */
